/*  VT1500 Device Driver for Linux, Version 1.0.10
 *  Copyright (C) 1995-2001  M. Gutschke
 *
 *  At the time of writing my e-mail address is:
 *	Internet: markus@gutschke.com
 *  My snail mail address is:
 *	Markus Gutschke
 *	3637 Fillmore Street #106
 *	San Francisco, CA 94123-1600
 *	USA
 *  If you like this software, I would appreciate if you sent me a postcard
 *  from your hometown. Under the terms of the GNU general public license
 *  you are free to include this program into (commercial) software
 *  distributions (e.g. putting it onto CD-ROM); nonetheless, I would really
 *  appreciate if you dropped me a short note (sending me a sample copy of
 *  your distribution would be even more appreciated!)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>

#include "osd.h"
#include "rc.h"
#include "server.h"
#include "tv.h"
#include "tvcontrol.h"

jmp_buf breakPressed;

       int    inetsock = 0;
static fd_set readfds;
static int    maxFd;
static int    chldCount;

static struct sockaddr *cur_addr;
static int    cur_fd,cur_data,cur_sock,cur_len;
static int    connection_closed;

static int __inline__ countFds(fd_set *fds,int *maxFd)
{
  int i,j,k;

  for (i = j = k = 0; i < FD_SETSIZE; i++)
    if (FD_ISSET(i,fds)) {j = i+1; k++;}
  if (maxFd) *maxFd = j;
  return(k);
}

static void servRequest(int sock)
{
  static const char ok[] = "OK\n";
  char   buf[80];
  int    resplen = 3;
  char   *response = (char *)ok;
  ReqRec req;
  int    rc,i;

  do {rc = read(sock,&req,sizeof(req));} while (rc < 0 && errno == EINTR);
  if (rc != sizeof(req)) {
    fprintf(stderr,"tv: Could not receive request: %s\n",
	    rc < 0 ? strerror(errno) : "read error");
    return; }
  switch (req.op) {
  case Onop:           break;
  case Ostr:           /* Not used; send echo message; for debugging only */
                       resplen = strlen(response = req.parm.str); break;
  case Otermnotify:    FD_CLR(req.data,&readfds); close(req.data);
                       /* do *not* send acknowledge */ return;
  case Okill:          exit(0);
  case Owarp:          {extern int warp(int);i = warp(req.parm.iParm);}
  iRet:                sprintf(response = buf,"%d\n",i);
                       resplen = strlen(buf); break;
  case OsetChannel:    i = setChannel(req.parm.iParm); goto iRet;
  case OsndCmd:        sndCmd(req.parm.iParm); break;
  case OsetRadio:      setRadio(req.parm.iParm); break;
  case OsetFrq:        setFrq(req.parm.iParm); break;
  case OsetVolume:     setVolume(req.parm.iParm); break;
  case OsetBass:       setBass(req.parm.iParm); break;
  case OsetTreble:     setTreble(req.parm.iParm); break;
  case OsetMute:       setMute(req.parm.iParm); break;
  case OtmpMute:       tmpMute(); break;
  case OrefreshMute:   refreshMute(); break;
  case OsetVGA:        setVGA(req.parm.iParm); break;
  case OsetTV:         setTV(); break;
  case OsetVCR:        setVCR(); break;
  case OsetPAL:        setPAL(); break;
  case OsetNTSC:       setNTSC(); break;
  case OsetHue:        setHue(req.parm.iParm); break;
  case OsetBrightness: setBrightness(req.parm.iParm); break;
  case OsetSaturation: setSaturation(req.parm.iParm); break;
  case OsetContrast:   setContrast(req.parm.iParm); break;
  case OsetRed:        setRed(req.parm.iParm); break;
  case OsetGreen:      setGreen(req.parm.iParm); break;
  case OsetBlue:       setBlue(req.parm.iParm); break;
  case OchannelName:   {int i;const char *j = channelName(req.parm.iParm,&i);
                       sprintf(response = buf,"%d,\"%s\"\n",i,j);
		       resplen = strlen(buf); break;}
  case OisVisible:     i = isVisible(); goto iRet;
  case OisRadio:       i = isRadio(); goto iRet;
  case OgetTVVCR:      i = getTVVCR(); goto iRet;
  case OgetPALNTSC:    i = getPALNTSC(); goto iRet;
  case OgetChannel:    i = getChannel(); goto iRet;
  case OgetFrq:        i = getFrq(); goto iRet;
  case OgetMute:       i = getMute(); goto iRet;
  case OgetExternalMute:   i = getExternalMute(); goto iRet;
  case OgetBass:       i = getBass(); goto iRet;
  case OgetTreble:     i = getTreble(); goto iRet;
  case OgetVolume:     i = getVolume(); goto iRet;
  case OgetInternalVolume: i = getInternalVolume(); goto iRet;
  case OgetExternalVolume: i = getExternalVolume(); goto iRet;
  case OgetBrightness: i = getBrightness(); goto iRet;
  case OgetSaturation: i = getSaturation(); goto iRet;
  case OgetContrast:   i = getContrast(); goto iRet;
  case OgetHue:        i = getHue(); goto iRet;
  case OgetRed:        i = getRed(); goto iRet;
  case OgetGreen:      i = getGreen(); goto iRet;
  case OgetBlue:       i = getBlue(); goto iRet;
  case OchannelInfo:   {const char *name,*longName;int frq,adj,prg,mode;
                       i = channelInfo(req.parm.iParms,req.parm.iParms[1],
				       req.parm.iParms[2],&name,&longName,
				       &frq,&adj,&prg,&mode);
                       sprintf(response = buf,"%d,%d,\"%s\",\"%s\",%d,%d,"
			       "%d,%d\n",i,req.parm.iParms[0],name,longName,
			       frq,adj,prg,mode);
		       resplen = strlen(buf); break;}
  case OsetChannelInfo:setChannelInfo(req.parm.x.iParms[0],req.parm.x.sParm,
				     req.parm.x.iParms[1],req.parm.x.iParms[2],
				     req.parm.x.iParms[3]); break;
  case OclearScreen:    clearScreen(); break;
  case OshowHelp:       showHelp(); break;
  case OshowTVVCR:      showTVVCR(req.parm.iParm); break;
  case OshowMute:       showMute(req.parm.iParm); break;
  case OshowBass:       showBass(req.parm.iParm); break;
  case OshowTreble:     showTreble(req.parm.iParm); break;
  case OshowVolume:     showVolume(req.parm.iParm); break;
  case OshowColor:      showColor(req.parm.iParms[0],req.parm.iParms[1],
				  req.parm.iParms[2],req.parm.iParms[3],
				  req.parm.iParms[4],req.parm.iParms[5],
				  req.parm.iParms[6],req.parm.iParms[7]);break;
  case OshowChannel:    showChannel(req.parm.y.iParms[0],req.parm.y.sParm1,
				    req.parm.y.sParm2); break;
  default:
    fprintf(stderr,"tv: Received illegal (0x%0X) request\n",req.op);
    break; }
  do {rc = write(req.data,response,resplen);
  } while (rc < resplen && errno == EINTR);
  if (rc != resplen) {
    fprintf(stderr,"tv: Could not send response: %s\n",
	    rc < resplen ? strerror(errno) : "write error");
    return; }
  return;
}

static void handleRequest(int fd,int data,ReqRec *req,int client_sock,
			  struct sockaddr *client_addr,int client_len)
{
  char   buffer[80];
  int    rc,len;

  req->data = data;
  do {rc = write(fd,req,sizeof(ReqRec));} while (rc < 0 && errno == EINTR);
  if (rc != sizeof(ReqRec))
    fprintf(stderr,"tv: Could not send request: %s\n",
	    rc < 0 ? strerror(errno) : "write error");
  do {
    do {rc = read(fd,buffer,sizeof(buffer));} while (rc < 0 && errno == EINTR);
    if (rc < 0)
      fprintf(stderr,"tv: Could not receive response: %s\n",strerror(errno));
    if (rc <= 0)
      return;
    len = rc;
    do {rc = write(client_sock,buffer,len);}while (rc < 0 && errno == EINTR);
    if (rc != len)
      fprintf(stderr,"tv: Could transmit response: %s\n",
	      rc < 0 ? strerror(errno) : "write error");
    if (ioctl(fd,FIONREAD,&rc)) {
      fprintf(stderr,"tv: FIONREAD failed: %s\n",strerror(errno));
      return; }
  } while (rc > 0);
  return;
}

int servread(void *buffer,int sz)
{
  int count;

  if (connection_closed)
    return(0);
  count = read(cur_sock,buffer,sz);
  if (count <= 0) {
    connection_closed = 1;
    return(0); }
  return(count);
}

void sendReq(ReqRec *req)
{
  handleRequest(cur_fd,cur_data,req,cur_sock,cur_addr,cur_len);
  return;
}

void writeClient(const char *buffer,int len)
{
  write(cur_sock,buffer,len);
  return;
}

static void doServer(int fd,int data,int client_sock,
		     struct sockaddr *client_addr,int client_len)
{
  static const char msg[] =
    "# VT1500 Device Driver for Linux, Version "VERSION"\n"
    "# Copyright (C) "YEAR" "AUTHOR"\n"
    "#\n"
    "# This program comes with ABSOLUTELY NO WARRANTY; for details\n"
    "# type `warranty();'. This is free software, and you are welcome\n"
    "# to redistribute it under certain conditions; type `conditions();'\n"
    "# for details. If you want to contact the author, enter `author()'.\n"
    "# For instructions enter `help();'\n"
    "\n";

  cur_fd   = fd;
  cur_data = data;
  cur_sock = client_sock;
  cur_addr = client_addr;
  cur_len  = client_len;
  connection_closed = 0;
  writeClient(msg,sizeof(msg)-1);
  setjmp(breakPressed);
  while (!connection_closed) {
    extern void servparse(void);
    servparse(); }
  return;
}

static void chldHandler(int signum)
{
  int rc,pid;

  UNUSED(signum);
  signal(SIGCHLD,chldHandler);
  chldCount--;
  pid = waitpid(0,&rc,0);
/*fprintf(stderr,"tv: Server died (pid = %d,%s = %d)\n",pid,
	  WIFSIGNALED(&rc) ? "sig" : "rc",
	  WIFSIGNALED(&rc) ? WTERMSIG(&rc) : WEXITSTATUS(&rc)); */
  return;
}

static void clientCleanup(void)
{
  ReqRec req;
  int    rc;

  req.data      = cur_data;
  req.op        = Otermnotify;
  do {rc = write(cur_fd,&req,sizeof(req));} while (rc < 0 && errno == EINTR);
  if (rc != sizeof(req))
    fprintf(stderr,"tv: Could not send termination notification: %s\n",
	    rc < 0 ? strerror(errno) : "write error");
  _exit(0); /* skip main program's atexit list */
}

static void serverCleanup(void)
{
  signal(SIGCHLD,chldHandler);
  signal(SIGTERM,SIG_IGN);
  shutdown(inetsock,2);
  close(inetsock);
  if (RCSettings.s.server.sin_family == AF_UNIX)
    unlink(RCSettings.s.server_unix.sun_path);
  kill(getpid(),SIGTERM);
  while (chldCount > 0 && sleep(2));
  return;
}

static void serverConnect(int inetsock)
{
  struct sockaddr_in client_addr;
  int client_sock,client_len = sizeof(client_addr);

retry:
  client_sock = accept(inetsock,(struct sockaddr *)&client_addr,&client_len);
  if (client_sock < 0) {
    if (errno == EINTR)
      goto retry;
    fprintf(stderr,"tv: Accept failure: %s\n",strerror(errno)); }
  else {
    int sockfds[2];
    int pid;

    if (socketpair(AF_UNIX,SOCK_DGRAM,0,sockfds)) {
      fprintf(stderr,"Could not open socket for server: %s\n",strerror(errno));
      exit(1); }
    signal(SIGCHLD,chldHandler);
    if ((pid = fork()) < 0) {
      fprintf(stderr,"Could not fork server: %s\n",strerror(errno));
      exit(1); }
    else if (pid == 0) {
      atexit(clientCleanup);
      close(sockfds[0]);
      doServer(sockfds[1],sockfds[0],client_sock,
	       (struct sockaddr *)&client_addr,client_len);
      exit(0); }
    else {
      close(client_sock);
      close(sockfds[1]);
      FD_SET(sockfds[0],&readfds); } }
  return;
}

int waitReadable(int fd,int delay)
{
  fd_set myfds;
  struct timeval tmo;
  int    rc;

  timerclear(&tmo); tmo.tv_sec = 5;
  for (;;) {
    myfds = readfds;
    FD_SET(fd,&myfds);
    chldCount = countFds(&myfds,&maxFd)-2;
    if ((rc = select(maxFd,&myfds,NULL,NULL,&tmo)) > 0) {
      if (FD_ISSET(inetsock,&myfds)) {
	serverConnect(inetsock); rc--; }
      if (FD_ISSET(fd,&myfds)) rc--;
      if (rc > 0) {
	int i;
	for (i = 3; rc > 0; i++)
	  if (FD_ISSET(i,&myfds)) {
	    rc--; servRequest(i); } }
      if (FD_ISSET(fd,&myfds))
	return(1); }
    else if (rc == 0)
      return(0); }
}

void initializeServer(void)
{
  struct linger      linger = {1,0};
  struct sockaddr    *serv_addr;
  int                addr_len;

  setpgid(0,0);
  FD_ZERO(&readfds);
  serv_addr = (struct sockaddr *)&RCSettings.s.server;
  addr_len  = serv_addr->sa_family == AF_UNIX ? sizeof(struct sockaddr_un)
                                              : sizeof(struct sockaddr_in);
  if ((inetsock = socket(serv_addr->sa_family,SOCK_STREAM,0)) < 0) {
    fprintf(stderr,"tv: Cannot open stream socket: %s\n",strerror(errno));
    exit(1); }
  if (setsockopt(inetsock,SOL_SOCKET,SO_LINGER,&linger,sizeof(linger)))
    fprintf(stderr,"tv: setsockopt(SO_LINGER) failed: %s\n",strerror(errno));
  if (setsockopt(inetsock,SOL_SOCKET,SO_REUSEADDR,serv_addr,addr_len))
    fprintf(stderr,"tv: setsockopt(SO_REUSEADDR) failed: %s\n",
	    strerror(errno));
  if (bind(inetsock,serv_addr,addr_len) < 0) {
    if (errno == EADDRINUSE) { /* switching to client mode */
      close(inetsock);
      inetsock = -1;
      return; }
    fprintf(stderr,"tv: Cannot bind local address: %s\n",strerror(errno));
    exit(1); }
  listen(inetsock,5);
  FD_SET(inetsock,&readfds);
  atexit(serverCleanup);
  return;
}
